import pytest
from heedy import App


def test_basics():
    c = App("gV9A4/J0eS4nLGAnzYyn")
    assert len(c.listObjects()) > 0


@pytest.mark.asyncio
async def test_basics_async():
    c = App("gV9A4/J0eS4nLGAnzYyn", session="async")
    assert len(await c.listObjects()) > 0
    print(await c.read())



import json
from urllib.parse import urljoin

# Used for the synchronous session
import requests

# Used for the asynchronous session
import aiohttp

DEFAULT_URL = "http://localhost:1324"


class HeedyError(Exception):
    """
    HeedyError is raised when the server returns an error value
    in response to a request.

    You can get the error contents by accessing the :code:"error" property
    and the :code:"error_description" property::

        # This is equivalent to print(myerror)
        print(f"{myerror.error}: {myerror.error_description}")
    """

    def __init__(self, msg):
        self.error = msg["error"]
        self.error_description = msg["error_description"]

    def __str__(self):
        return f"{self.error}: {self.error_description}"


class Session:
    """
    Session is the abstract base class that both sync and async sessions implement
    """

    def __init__(self, url=DEFAULT_URL):

        # Set up the API url
        if not url.startswith("http"):
            url = "https://" + url
        if not url.endswith("/"):
            url = url + "/"
        self.url = url

    def setAccessToken(self, token):
        raise NotImplementedError()

    def setPluginKey(self, key):
        raise NotImplementedError()

    def version(self):
        raise NotImplementedError()

    def get(self, path, params={}, f=lambda x: x):
        raise NotImplementedError()

    def post(self, path, data, params={}, f=lambda x: x):
        raise NotImplementedError()

    def patch(self, path, data, params={}, f=lambda x: x):
        raise NotImplementedError()

    def delete(self, path, params={}, f=lambda x: x):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class SyncSession(Session):
    """
    SyncSession is to be used in synchronous programs. It uses requests internally.
    """

    def __init__(self, url=DEFAULT_URL):
        super().__init__(url)
        self.s = requests.Session()
        self.s.headers.update({'Content-Type': 'application/json'})

    def setAccessToken(self, token):
        self.s.headers.update({'Authorization': f"Bearer {token}"})

    def setPluginKey(self, key):
        self.s.headers.update({"X-Heedy-Key": key})

    def version(self):
        return self.handleResponse(self.s.get(urljoin(self.url, "api/heedy/v1/server/version"))).text

    def handleResponse(self, r):
        if r.status_code >= 400:
            # The response returned an error
            try:
                msg = r.json()
            except:
                msg = {"error": "malformed_response", "error_description":
                       f"The server returned \"{r.text}\", which is not json."}
            raise HeedyError(msg)
        return r

    def get(self, path, params={}, f=lambda x: x):
        return f(self.handleResponse(self.s.get(urljoin(self.url, path), params=params)).json())

    def post(self, path, data, params={}, f=lambda x: x):
        return f(self.handleResponse(self.s.post(urljoin(self.url, path), data=json.dumps(data), params=params)).json())

    def patch(self, path, data, params={}, f=lambda x: x):
        return f(self.handleResponse(self.s.patch(urljoin(self.url, path), data=json.dumps(data), params=params)).json())

    def delete(self, path, params={}, f=lambda x: x):
        return f(self.handleResponse(self.s.delete(urljoin(self.url, path), params=params)).json())

    def close(self):
        self.s.close()


class AsyncSession(Session):
    """
    AsyncSession is used when running in an asyncio event loop. All of the requests become coroutines,
    allowing them to be awaited
    """

    def __init__(self, url=DEFAULT_URL):
        super().__init__(url)
        self.s = None
        self.headers = {'Content-Type': 'application/json'}

    def setAccessToken(self, token):
        self.headers['Authorization'] = f"Bearer {token}"

    def setPluginKey(self, key):
        self.headers["X-Heedy-Key"] = key

    def initSession(self):
        if self.s is None:
            self.s = aiohttp.ClientSession()

    async def handleResponse(self, r):
        if r.status >= 400:
            # The response returned an error
            try:
                msg = await r.json()
            except:
                msg = {"error": "malformed_response",
                       "error_description": f"The server did not return valid json"}
            raise HeedyError(msg)
        return r

    async def version(self):
        self.initSession()
        return await (await self.handleResponse(await self.s.get(urljoin(self.url, "api/heedy/v1/server/version")))).text()

    async def get(self, path, params={}, f=lambda x: x):
        self.initSession()
        return f(await (await self.handleResponse(await self.s.get(urljoin(self.url, path), params=params, headers=self.headers))).json())

    async def post(self, path, data, params={}, f=lambda x: x):
        self.initSession()
        return f(await (await self.handleResponse(await self.s.post(urljoin(self.url, path), params=params, data=json.dumps(data), headers=self.headers))).json())

    async def patch(self, path, data, params={}, f=lambda x: x):
        self.initSession()
        return f(await (await self.handleResponse(await self.s.patch(urljoin(self.url, path), params=params, data=json.dumps(data), headers=self.headers))).json())

    async def delete(self, path, params={}, f=lambda x: x):
        self.initSession()
        return f(await (await self.handleResponse(await self.s.delete(urljoin(self.url, path), params=params, headers=self.headers))).json())

    async def raw(self, method, path, data=None, params={}, headers={}):
        self.initSession()
        return await self.s.request(method, urljoin(self.url, path), headers={**self.headers, **headers}, data=data)

    async def close(self):
        if self.s is not None:
            await self.s.close()


def getSessionType(sessionType: str, url: str = DEFAULT_URL) -> Session:
    """
    This function is given a string, either "sync" or "async", and it returns a SyncSession or AsyncSession respectively.
    """
    if sessionType == "sync":
        return SyncSession(url)
    if sessionType == "async":
        return AsyncSession(url)
    raise NotImplementedError(
        f"The session type '{sessionType}' is not implemented")


class APIObject:
    """
    APIObject represents an object in heedy (user,app,object,etc).
    It is given a session and the api location of the object, and allows
    reading, updating, and deleting the object
    """

    def __init__(self, session: Session, uri: str):
        self.session = session
        self.uri = uri

    def read(self, **kwargs):
        """
        Read the object
        """
        return self.session.get(self.uri, params=kwargs)

    def update(self, **kwargs):
        """
        Updates the given data::

            o.update(name="My new name",description="my new description")
        """
        return self.session.patch(self.uri, kwargs)

    def delete(self, **kwargs):
        """
        Deletes the object
        """
        return self.session.delete(self.uri, params=kwargs)



from .base import APIObject, Session
from typing import Dict


class Object(APIObject):
    def __init__(self, session: Session, objectData: Dict):
        super().__init__(session, f"api/heedy/v1/objects/{objectData['id']}")
        self.data = objectData

    @property
    def type(self) -> str:
        """
        Returns the object type
        """
        return self.data["type"]

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return str(self)


_objectRegistry = {}


def registerObjectType(objectType: str, objectClass: Object) -> None:
    """
    registerObjectType allows external libraries to implement object types
    available through heedy plugins. All you need to do is subclass :code:`Object`,
    and register the corresponding type!
    """
    _objectRegistry[objectType] = objectClass


def getObjectObject(session: Session, objectData: Dict) -> Object:
    """
    Heedy allows multiple different object types. getObjectObject uses the
    registered object type objects to initialize the given object data to
    the correct object type. If the object is of an unregistered type, it returns
    a base :code:`Object` object.
    """
    if objectData["type"] in _objectRegistry:
        return _objectRegistry[objectData["type"]](session, objectData)
    return Object(session, objectData)



from .base import getSessionType
import json
import os
import multidict
import aiohttp
import base64


class Plugin:
    def __init__(self, session: str = "async"):
        # Load the plugin configuration
        self.config = json.loads(input())

        # Change the directory to the data dir
        os.chdir(self.config["data_dir"])

        self.session = getSessionType(
            session, f"http://localhost:{self.config['config']['port']}")
        self.session.setPluginKey(self.config["apikey"])

    @property
    def name(self):
        return self.config["plugin"]

    async def forward(self, request, data=None, headers={}, run_as: str = None, overlay=None):
        """
        Forwards the given request to the underlying database.
        It only functions in async mode.

        Returns the response.
        """
        if data is None:
            data = await request.read()
        headers = {**request.headers, **headers}

        if overlay is None:
            overlay = "next"

        headers["X-Heedy-Overlay"] = overlay

        return await self.session.raw(request.method, request.path, headers=headers, data=data, params=request.query)

    async def respond_forwarded(self, request, **kwargs):
        """
        Responds to the request with the result of forward()
        """
        req_res = await self.forward(request, **kwargs)

        response = aiohttp.web.StreamResponse(
            status=req_res.status, headers=req_res.headers)
        await response.prepare(request)
        while True:
            chunk = await req_res.content.read(32768)
            if not chunk:
                break
            await response.write(chunk)
        await response.write_eof()
        return response

    def objectRequest(self, request):
        h = request.headers
        last_modified = h["X-Heedy-Last-Modified"]
        if last_modified == "null":
            last_modified = None
        return {
            "request": h["X-Heedy-Request"],
            "id": h["X-Heedy-Id"],
            "last_modified": last_modified,
            "meta": json.loads(base64.b64decode(h["X-Heedy-Meta"])),
            "object": h["X-Heedy-Object"],
            "owner": h["X-Heedy-Owner"],
            "as": h["X-Heedy-As"],
            "access": h["X-Heedy-Access"].split(" ")
        }

    def hasAccess(self, request, scope):
        a = request.headers["X-Heedy-Access"].split(" ")
        return "*" in a or scope in a

    def isUser(self, request):
        h = request.headers
        return not "/" in h["X-Heedy-As"] and h["X-Heedy-As"] != "heedy" and h["X-Heedy-As"] != "public"

    def isApp(self, request):
        return "/" in request.headers["X-Heedy-As"]

    def isAdmin(self, request):
        request.headers["X-Heedy-As"] == "heedy"

    def fire(self, event):
        """
        Fires the given event
        """
        return self.session.post("/api/heedy/v1/events", event)

    def listApps(self, **kwargs):
        return self.session.get("api/heedy/v1/apps", params=kwargs)

    def notify(self, n, **kwargs):
        return self.session.post("/api/heedy/v1/notifications", n, params=kwargs)

    def delete_notification(self, n):
        return self.session.delete("/api/heedy/v1/notifications", params=n)



from .base import APIObject, getSessionType, DEFAULT_URL
from .object import getObjectObject

from functools import partial


class App(APIObject):
    def __init__(self, access_token: str, url: str = DEFAULT_URL, session: str = "sync"):
        # Initialize the app object
        s = getSessionType(session)
        s.setAccessToken(access_token)
        super().__init__(s, "api/heedy/v1/apps/self")

    def createObject(self, **kwargs):
        return self.session.post("api/heedy/v1/objects", data=kwargs, f=lambda x: getObjectObject(self.session, x))

    def listObjects(self, **kwargs):
        return self.session.get("api/heedy/v1/objects",
                                params=kwargs, f=lambda x: list(map(partial(getObjectObject, self.session), x)))

    def notify(self, n, **kwargs):
        return self.session.post("/api/heedy/v1/notifications", n, params=kwargs)
